"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Highlight_1 = require("./Highlight");
const injectHighlightWrappers_1 = require("./injectHighlightWrappers");
const rangeContents_1 = require("./rangeContents");
const removeHighlightWrappers_1 = require("./removeHighlightWrappers");
const selection_1 = require("./selection");
const SerializedHighlight_1 = require("./SerializedHighlight");
class Highlighter {
    constructor(container, options = {}) {
        this.highlights = {};
        this.eraseAll = () => {
            this.getHighlights().forEach(this.erase);
        };
        this.erase = (highlight) => {
            removeHighlightWrappers_1.default(highlight);
            delete this.highlights[highlight.id];
        };
        this.onMouseup = () => {
            const selection = this.document.getSelection();
            if (!selection) {
                return;
            }
            if (selection.isCollapsed) {
                this.onClick(selection);
            }
            else {
                this.onSelect(selection);
            }
        };
        this.container = container;
        this.options = options;
        this.container.addEventListener('mouseup', this.onMouseup);
    }
    unmount() {
        this.container.removeEventListener('mouseup', this.onMouseup);
    }
    highlight(highlight) {
        if (highlight instanceof SerializedHighlight_1.default && highlight.isLoadable(this)) {
            return this.highlight(highlight.load(this));
        }
        else if (highlight instanceof Highlight_1.default) {
            this.highlights[highlight.id] = highlight;
            injectHighlightWrappers_1.default(highlight, this.options);
        }
    }
    getHighlight(id) {
        return this.highlights[id];
    }
    getReferenceElement(id) {
        return this.container.querySelector(`[id="${id}"]`);
    }
    clearFocus() {
        this.container.querySelectorAll(`.${this.options.className}.${Highlight_1.FOCUS_CSS}`)
            .forEach((el) => el.classList.remove(Highlight_1.FOCUS_CSS));
    }
    getHighlights() {
        const highlights = Object.values(this.highlights);
        highlights.sort((a, b) => {
            return a.range.compareBoundaryPoints(Range.START_TO_START, b.range);
        });
        return highlights;
    }
    get document() {
        if (!this.container.ownerDocument) {
            throw new Error('highlighter container is not mounted to a document!');
        }
        return this.container.ownerDocument;
    }
    onClick(selection) {
        const { onClick } = this.options;
        const clickedHighlight = () => {
            if (selection.rangeCount < 1) {
                return;
            }
            const range = selection_1.getRange(selection);
            return Object.values(this.highlights)
                .find((other) => other.intersects(range));
        };
        if (onClick) {
            onClick(clickedHighlight());
        }
    }
    onSelect(selection) {
        const { onSelect } = this.options;
        const range = selection_1.snapSelection(selection, this.options);
        if (onSelect && range) {
            const highlights = Object.values(this.highlights)
                .filter((other) => other.intersects(range));
            if (highlights.length === 0) {
                const highlight = new Highlight_1.default(range, { content: rangeContents_1.rangeContentsString(range) });
                onSelect(highlights, highlight);
            }
            else {
                onSelect(highlights);
            }
        }
    }
}
exports.default = Highlighter;
//# sourceMappingURL=Highlighter.js.map